import os
import sys
from functools import wraps, partial
from contextlib import contextmanager
from uuid import uuid1
from Queue import Empty
from multiprocessing import (Queue,
                             active_children,
                             Lock,
                             cpu_count,
                             connection,
                             current_process
                             )
from types import StringTypes

from multiprocessing.util import Finalize, _finalizer_counter, _finalizer_registry
from multiprocessing.managers import dispatch, State, BaseProxy
import multiprocessing

import time

from traceback import print_exc
from werkzeug.datastructures import ImmutableDict
from werkzeug.routing import Map, Rule, BuildError
from werkzeug.utils import import_string, cached_property



from .globals import _swarm_ctx_stack
from .config import ConfigAttribute, Config
from .context import SwarmContext
from .helpers import (_PackageBoundObject,
                      find_package,
                      locked_cached_property,
                      _endpoint_from_swarm_func,
                      cascade_reload,
                      obj_converter,
                      obj_list_converter
                      )
from .exceptions import Cancel
from .logging import create_app_logger, create_context_logger
from swarm import signals
from .state import save_state


import logging


ident = lambda:(current_process().pid, id(threading.current_thread()))
from pprint import pprint
from traceback import print_exc
import time

_logger_lock = multiprocessing.Lock()

class BaseSwarm(_PackageBoundObject):
    __OMIT = ('_log', '_log_handler')
    def __getstate__(self):
        return dict((k, v) for (k, v) in self.__dict__.items() if not k in self.__class__.__OMIT)

    #: Default configuration parameters.
    blocks = {}
    default_config = ImmutableDict({
        'DEBUG': False,
        'COMPETITORS': cpu_count(),
        'BIND_HOST': '127.0.0.1',
        'PATH_INFO': '/',
        'RESOLUTION': 0.01,
        'SWARM_CONTEXT_CLASS': 'swarm.context.SwarmContext',
        'PROCESS_CONTEXT_CLASS': 'swarm.context.ProcessContext',
        'TRANSPORT_CLASS': 'swarm.app.BaseTransport',
        'DEFAULT_METHOD':'GET',
        'LOGGER_NAME': None,
        'LOG_LEVEL': 'logging.ERROR',
        'LOG_FORMAT': '%(asctime)s:%(msecs)-3d %(processName)-24s(%(process)-5s):%(name)-22s %(levelname)-10s - %(message)s',
        'RELOADED_MODULE':None,
        'SAVE_STATE': False,
    })

    #: The rule object to use for URL rules created.  This is used by
    #: :meth:`add_url_rule`.  Defaults to :class:`werkzeug.routing.Rule`.
    #:
    #: .. versionadded:: 0.1
    url_rule_class = Rule
    
    logger_name = ConfigAttribute('LOGGER_NAME')
    log_level = ConfigAttribute('LOG_LEVEL', get_converter= obj_converter)
    log_format = ConfigAttribute('LOG_FORMAT')
    resolution = ConfigAttribute('RESOLUTION')
    debug = ConfigAttribute('DEBUG')

    transport_class = ConfigAttribute('TRANSPORT_CLASS',
                                      get_converter= obj_converter)

    swarm_context_class = ConfigAttribute('SWARM_CONTEXT_CLASS',
                                      get_converter= obj_converter)

    process_context_class = ConfigAttribute('PROCESS_CONTEXT_CLASS',
                                      get_converter= obj_converter)


    competitors =  ConfigAttribute('COMPETITORS')

    reloaded_module = ConfigAttribute('RELOADED_MODULE', get_converter=obj_list_converter)

    def __init__(self, import_name, static_path=None,
                 persistance_folder='persist',
                 instance_path=None,
                 instance_relative_config=False,
                 log_handler=None):

        _PackageBoundObject.__init__(self, import_name)

        if persistance_folder is not None:
            self.persistance_folder = persistance_folder
        if instance_path is None:
            instance_path = self.auto_find_instance_path()
        elif not os.path.isabs(instance_path):
            raise ValueError('If an instance path is provided it must be '
                             'absolute.  A relative path was given instead.')

        #: Holds the path to the instance folder.
        #:
        #: .. versionadded:: 0.1
        self.instance_path = instance_path

        #: The configuration dictionary as :class:`Config`.  This behaves
        #: exactly like a regular dictionary but supports additional methods
        #: to load a config from files.
        self.config = self.make_config(instance_relative_config)

        # Prepare the deferred setup of the logger.
        self._logger = None

        #: A dictionary of all view functions registered.  The keys will
        #: be function names which are also used to generate URLs and
        #: the values are the function objects themselves.
        #: To register a view function, use the :meth:`route` decorator.
        self.swarm_functions = {}

        #: The :class:`~werkzeug.routing.Map` for this instance.  You can use
        #: this to change the routing converters after the class was created
        #: but before any routes are connected.  Example::
        #:
        #:    from werkzeug.routing import BaseConverter
        #:
        #:    class ListConverter(BaseConverter):
        #:        def to_python(self, value):
        #:            return value.split(',')
        #:        def to_url(self, values):
        #:            return ','.join(BaseConverter.to_url(value)
        #:                            for value in values)
        #:
        #:    app = Flask(__name__)
        #:    app.url_map.converters['list'] = ListConverter
        self.url_map = Map()
        self._log_handler = log_handler
    
    def add_url_rule(self, rule, endpoint=None, swarm_func=None, **options):
        """Connects a URL rule.  Works exactly like the :meth:`route`
        decorator.  If a subprocess_func is provided it will be registered with the
        endpoint.

        Basically this example::

            @app.route('/')
            def index():
                pass

        Is equivalent to the following::

            def index():
                pass
            app.add_url_rule('/', 'index', index)

        If the subprocess_func is not provided you will need to connect endpoint
        to a view function like so::

            app.subprocess_functions['index'] = index

        Internally :meth:`route` invokes :meth:`add_url_rule` so if you want
        to customize the behavior via subclassing you only need to change
        this method.

        :param rule: the URL rule as string
        :param endpoint: the endpoint for the registered URL rule.  Flask
                         itself assumes the name of the view function as
                         endpoint
        :param subprocess_func: the function to call when serving a request to the
                          provided endpoint
        :param options: the options to be forwarded to the underlying
                        :class:`~werkzeug.routing.Rule` object.  A change
                        to Werkzeug is handling of method options.  methods
                        is a list of methods this rule should be limited
                        to (`GET`, `POST` etc.).  By default a rule
                        just listens for `GET` (and implicitly `HEAD`).
                        Starting with Flask 0.6, `OPTIONS` is implicitly
                        added and handled by the standard request handling.
        """
        if endpoint is None:
            endpoint = _endpoint_from_swarm_func(swarm_func)
        options['endpoint'] = endpoint

        methods = options.pop('methods', None)

        # if the methods are not given and the view_func object knows its
        # methods we can use that instead.  If neither exists, we go with
        # a tuple of only `GET` as default.
        if methods is None:
            methods = getattr(swarm_func, 'methods', None) or ('GET',)
        methods = set(methods)

        # Methods that should always be added
        required_methods = set(getattr(swarm_func, 'required_methods', ()))

        # starting with Flask 0.8 the swarm_func object can disable and
        # force-enable the automatic options handling.
        provide_automatic_options = getattr(swarm_func,
                                            'provide_automatic_options', None)

        if provide_automatic_options is None:
            if 'OPTIONS' not in methods:
                provide_automatic_options = True
                required_methods.add('OPTIONS')
            else:
                provide_automatic_options = False

        # Add the required methods now.
        methods |= required_methods

        # starting with Flask 0.8 the swarm_func object can disable and
        # force-enable the automatic options handling.
        # due to a werkzeug bug we need to make sure that the defaults are
        # None if they are an empty dictionary.  This should not be necessary
        # with Werkzeug 0.7
        options['defaults'] = options.get('defaults') or None

        rule = self.url_rule_class(rule, methods=methods, **options)
        self.url_map.add(rule)

        if swarm_func is not None:
            self.swarm_functions[endpoint] = swarm_func

    def endpoint(self, endpoint):
        """A decorator to register a function as an endpoint.
        Example::

            @app.endpoint('example.endpoint')
            def example():
                return "example"

        :param endpoint: the name of the endpoint
        """
        def decorator(f):
            self.swarm_functions[endpoint] = f
            return f
        return decorator

    def make_config(self, instance_relative=False):
        """Used to create the config attribute by the Flask constructor.
        The `instance_relative` parameter is passed in from the constructor
        of Flask (there named `instance_relative_config`) and indicates if
        the config should be relative to the instance path or the root path
        of the application.

        .. versionadded:: 0.1
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        return Config(root_path, self.default_config)

    def auto_find_instance_path(self):
        """Tries to locate the instance path if it was not provided to the
        constructor of the application class.  It will basically calculate
        the path to a folder named ``instance`` next to your main file or
        the package.

        .. versionadded:: 0.1
        """
        prefix, package_path = find_package(self.import_name)
        if prefix is None:
            return os.path.join(package_path, 'instance')
        return os.path.join(prefix, 'var', self.name + '-instance')

    def open_instance_resource(self, resource, mode='rb'):
        """Opens a resource from the application's instance folder
        (:attr:`instance_path`).  Otherwise works like
        :meth:`open_resource`.  Instance resources can also be opened for
        writing.

        :param resource: the name of the resource.  To access resources within
                         subfolders use forward slashes as separator.
        """
        return open(os.path.join(self.instance_path, resource), mode)

    @locked_cached_property
    def name(self):
        """The name of the application.  This is usually the import name
        with the difference that it's guessed from the run file if the
        import name is main.  This name is used as a display name when
        Swarm needs the name of the application.  It can be set and overriden
        to change the value.
        """
        if self.import_name == '__main__':
            fn = getattr(sys.modules['__main__'], '__file__', None)
            if fn is None:
                return '__main__'
            return os.path.splitext(os.path.basename(fn))[0]
        return self.import_name

    @cached_property
    def log(self):
        if self.config['DEBUG']:
            self.log_level = logging.DEBUG
        if len(_swarm_ctx_stack) > 1:
            return create_context_logger(_swarm_ctx_stack.bottom,
                                         _swarm_ctx_stack.top._log_queue)
        else:
            return create_app_logger(self)

    def log_handler(self, *args, **kwargs):
        return logging.StreamHandler(stream=sys.stdout)
        if not getattr(self,  '_log_handler', None):
            self._log_handler = logging.StreamHandler(stream=sys.stdout)
        return  self._log_handler
    
    @locked_cached_property
    def url_adapter(self):
        return self.url_map.bind(self.config['BIND_HOST'],
                                 path_info=self.config['PATH_INFO'])

    def url(self, rule, **options):
        """A decorator that is used to register a subprocess function for a
        given URL rule.  This does the same thing as :meth:`add_url_rule`
        but is intended for decorator usage::

            @app.route('/')
            def index():
                return 'Hello World'

        For more information refer to :ref:`url-route-registrations`.

        :param rule: the URL rule as string
        :param endpoint: the endpoint for the registered URL rule.  Flask
                         itself assumes the name of the view function as
                         endpoint
        :param swarm_func: the function to call when serving a request to the
                          provided endpoint
        :param options: the options to be forwarded to the underlying
                        :class:`~werkzeug.routing.Rule` object.  A change
                        to Werkzeug is handling of method options.
        """
        def decorator(f):
            endpoint = options.pop('endpoint', None)
            self.add_url_rule(rule, endpoint, f, **options)
            return f
        return decorator        

    @staticmethod
    def _defer_shutdown(caller, calee, context, log):
        log.debug('Waiting for manager to send state before shutdown')
        calee.send(None)

        tasks = []
        task = calee.recv()
        while task is not None:
            tasks += task,
            task = calee.recv()
        caller.close()
        calee.close()
        log.debug('Saving %s tasks'%len(tasks))
        items = []
        while True:
            try:
                items += context.items.get_nowait(),
            except Empty:
                break
        #somewhy first get is always empty
        while True:
            try:
                items += context.items.get_nowait(),
            except Empty:
                break

        log.debug('Saving %s items'%len(items))
        save_state(context, tasks=tasks, items=items)
        log.info('Shutting down with id %s'%context._process.pid)
        if 'log' in context.object.__dict__ and hasattr(context.object.log, 'close'):
            try:
                context.object.log.close()
            except Empty:
                pass
            del context.object.__dict__['log']
        if items or tasks:
            context.object.pent.append(context._process.pid)


    def defer_shutdown(self, caller, calee, context):
        if not self.config['SAVE_STATE']:
            return
        context._defer_shutdown = Finalize(context,
                                          type(self)._defer_shutdown,
                                          args=(caller, calee, context, self.log), 
                                          exitpriority=11)


    @contextmanager
    def context(self, caller, calee, args, init='_init'):

        if self.reloaded_module:
            cascade_reload(self.reloaded_module)

        context = self.swarm_context_class(self,
                                           caller,
                                           *args)

        is_top = not len(_swarm_ctx_stack)
        if is_top:
            signals.swarm.signal('enter-top-context').send(self,
                                                           context=context)
        _swarm_ctx_stack.push(context)
        try:
            context.start(getattr(context, init), ignore_exit=self.config['SAVE_STATE'])

            if is_top:
                self.log.recieve(calee.recv())
            else:
                calee.recv()
            
            self.defer_shutdown(caller, calee, context)

            context.items = calee.recv()
            context.pool_is_running = calee.recv()
            context.pool_counter_lock = calee.recv()
            signals.swarm.signal('enter-context').send(self, context=context)

            yield context

        finally:
            context = _swarm_ctx_stack.pop()
            signals.swarm.signal('exit-context').send(self,
                                                       context=context)
            if is_top:
                signals.swarm.signal('exit-top-context').send(self,
                                                                context=context)

            if hasattr(context, '_defer_shutdown'):
                context._defer_shutdown()

            if 'log' in self.__dict__ and hasattr(self.log, 'close'):
                self.log.close()
                del self.__dict__['log']
            signals.disconnect()
            context.shutdown()

    @cached_property
    def pent(self):
        return []

    def iter_items(self, context):
        while True:
            try:
                item = context.items.get(True, self.resolution)
                self.log.debug('Got item %s'%type(item))
                yield item
            except Empty:
                with context.pool_counter_lock:
                    if not context.pool_is_running.value:
                        break
            except (KeyboardInterrupt, SystemExit):
                return



    def __call__(self, url, *args, **kwargs):
        """
        locks until call finished (has active_children)
        Usage:

        """
        caller, calee = connection.Pipe()
        #protocol:
        #calee->caller: log queue Queue()
        #calee->caller: items Queue()
        #calee->caller: pool tasks counter Value()
        #calee->caller: pool tasks counter Lock()
        #caller->calee: url, kwargs
        #calee->caller: None (block exec finished, wait for items, check pool_is_running)
        #caller->calee: None (Close pool execution and send all left tasks to this pipe)
        #calee->caller: [tasks,]
        #calee->caller: None
        #close

        try:
            with self.context(caller, calee, args) as context:
                self.log.info('Entered context: %s'%url + ' (**%s)'%kwargs if kwargs else '')
                calee.send((url, kwargs))
                calee.recv()
                for i in self.iter_items(context):
                    yield i
                self.log.info('Exited context: %s'%url)
        except Cancel, e:
            self.log.log(e.level, e.message%{'signal':e.signal.name})

    def restore(self, _id, **kwargs):
        caller, calee = connection.Pipe()
        with self.context(caller, calee, kwargs, init='_restore') as context:
            self.log.info('Start restoring: %s'%_id)
            calee.send(_id)
            calee.recv()
            for i in self.iter_items(context):
                yield i
            self.log.info('Finish restoring: %s'%_id)

    def instance_dir(self, name):
        _dir = os.path.join(self.instance_path, name)
        if not os.path.exists(_dir):
            os.makedirs(_dir)
        return _dir

class BaseTransport(object):
    def __init__(self, context, *args, **kwargs):
        self.context = context
        self.args = args
        self.kwargs = kwargs

    def perform(self):
        self.context.log.warning('BaseTransport.perform() called')
        return self

    @classmethod
    def iter_tasks(self, task, *args, **kwargs):
        for arg in args:
            yield task, (arg,), kwargs