#
# Name: codeblocks.py
#
# Synopsis: codeblocks v. 0.1b
#
# Description: This module lets you use code blocks in your code.
#
# Copyright 2012 Massimiliano Tomassoli (kiuhnm03@yahoo.it)
#
#                          The Wide Open License (WOL)
#
# Permission to use, copy, modify, distribute and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice and this license appear in all source copies.
# THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY OF
# ANY KIND. See http://www.dspguru.com/wide-open-license for more information.
#

'''
Usage
-----

Just import this module and call rewrite at the beginning of your own module,
i.e. before any other statement.
If your module is not the main module (__name__ != __main__), you must also
call end_of_module at the end of your module. An exception will be generated if
you forget to do that.

How it works
------------

This module loads the source code of your module, rewrites it by following some
rewriting-rule described below, executes it, and skips the original code (that's
really an implementation detail).

Implementation details:
    - The code of the main module is skipped by calling quit().
    - The code of a non-main module is skipped by jumping to the line marked by
      the call to end_of_module.

Bug in Python
-------------

Python has a bug (http://bugs.python.org/issue14612) which makes the function
skip_module_code crash if one or more with statements appear at the module level
in an imported module which imports codeblocks and calls rewrite.
For now, I've added a check to avoid the occurrence of such a situation.

Exceptions
----------

Exceptions are generated, in this order, in these phases:
    1) Initial parsing (performed by the interpreter itself).
    2) Parsing of code blocks (performed by this module).
    3) Compilation of the rewritten code (controlled by this module)
    4) Execution of the rewritten code (controlled by this module)

All the exceptions are signaled as if raised by /your/ code.

For exceptions raised during phases 2 and 3, the stack will show an additional
frame referring to your code (as in the other phases), but the error message
will be duplicated. Here's an example:

--->
Traceback (most recent call last):
  File "C:\path\test.py", line 10, in <module>
    codeblocks.rewrite()
  File "C:\path\test.py", line 583, in rewrite
    exec(compile(src, code_info.fname, 'exec'))
  File "C:\path\test.py\test.py", line 163, in <module>
    _return(4)
  File "C:\path\test.py\test.py", line 163
    _return(4)
    ^
SyntaxError: 'return' outside function
<---

Your debugger should break and point to line 163 of test.py.

Implementation details:
    - The duplication of the last error message seems to be unavoidable if we
      want both to show the arrow (^) and to force debuggers to break on the
      the right line of the source code.

Debugging
---------

Stepping through the code is fully supported. Original and rewritten code are
"synchronized" so that everything works as expected.

Example:

--->
1:  def each(iterable, block):
2:      for e in iterable:
3:          block(e)                        # step into -> 6:
4:                            
5:  with each(range(0, 10)) << 'x':         # step into -> 2:
6:      print('element ' + str(x))
<---

Issuing the "step into" command at line 3 takes you to line 6 while doing that
at line 5 takes you to line 2.
"Step over" commands also behave as expected. For instance, issuing a step over
command at line 5 will jump to the end of the code.

With-rewriting rules
--------------------

Code blocks are always passed to a single function.
Four forms are allowed:

1)  'with' [ret_vars <<] func(func_args) [<< codekw] << code_args
        body

2)  'with' [ret_vars <<] func(func_args) [<< codekw] << ':list':
        'with' int_code_args_1:
            body_1
        ...
        'with' int_code_args_n:
            body_n

3)  'with' [ret_vars <<] func(func_args) [<< codekw] << ':dict':
        'with' << name_1 << int_code_args_1:
            body_1
        ...
        'with' << name_n << int_code_args_n:
            body_n

4)  'with' [ret_vars <<] func(func_args) << ':multi'
        'with' [<< int_codekw_1] << int_code_args_1:
            body_1
        ...
        'with' [<< int_codekw_n] << int_code_args_n:
            body_n

where code_args is not in (':dict', ':list').

ret_vars, func_args, codekw and int_codekw_1, ..., int_codekw_n can be literals
or expressions (identifiers, tuples, etc...). code_args and int_code_args_1, ...,
int_code_args_n must be literals but with one exception: when the block they are
referring to takes no arguments, they must be _, None or ''.

Form 1 is rewritten as
    def unique_name(code_args):
        rew(body)
    [ret_vars =] func(func_args, [codekw =] unique_name)
where:
- 'func_args, [codekw =] unique_name' is func_args with the addition of
  unique_name as a keyword argument with name codekw if codekw is present, or
  as a positional argument if codekw is not present.
  Note that, in the real implementation, positional and keyword arguments will
  be inserted in the right position.
- rew(body) is the body recursively rewritten (body could contain other with
  statements which need to be rewritten.
- codekw can be an /arbitrary/ string, if it is a literal.

Analoguous remarks apply to the following forms.

Form 2 is rewritten as
    def unique_name_1(int_code_args_1):
        rew(body_1)
    ...
    def unique_name_n(int_code_args_n):
        rew(body_n)
    [ret_vars =] func(func_args, [codekw =] list(unique_name_1, ... unique_name_n))

Form 3 is rewritten as
    def unique_name_1(int_code_args_1):
        rew(body_1)
    ...
    def unique_name_n(int_code_args_n):
        rew(body_n)
    [ret_vars =] func(func_args,
                      [codekw =] {name_1 : unique_name_1,
                                  ...,
                                  name_n : unique_name_n})

Form 4 is rewritten as
    def unique_name_1(int_code_args_1):
        rew(body1)
    ...
    def unique_name_n(int_code_args_n):
        rew(bodyn)
    [ret_vars =] func(func_args,
                      [int_codekw_1 =] unique_name_1,
                      ...,
                      [int_codekw_n =] unique_name_n)

Expression-rewriting rules
--------------------------

In order to avoid some syntax errors, the following rewriting rules were added:

_return(expr)   --->   return expr
_yield(expr)    --->   yield expr
_global(ids)    --->   global ids
_nonlocal(ids)  --->   nonlocal ids
'''

import ast
import inspect
from types import ModuleType
import sys
import re
import traceback

# # Use threading only if available.

# try:
#     import threading as _threading
# except ImportError:
#     import dummy_threading as _threading

from multiprocessing import Lock

class _Reraise(Exception):
    ''' Preserves exception data for later use. '''
    def __init__(self, exc_type, msg, lineno, col_offset, code_info):
        self.__data = exc_type, msg, lineno, col_offset, code_info

    def get_data(self):
        return self.__data

def _raise(exc_type, msg, lineno, col_offset, code_info):
    ''' Simple wrapper around _Reraise.__init__() '''
    raise _Reraise(exc_type, msg, lineno, col_offset, code_info)

class EndOfModuleError(SyntaxError):
    pass
BLOCK_PREFIX = 'block_'
class _UniqueNames:
    ''' Generates unique names for anonymous block-functions '''
    __counter = 0
    __lock = Lock()
    __name_prefix = BLOCK_PREFIX

    @staticmethod
    def get_name():
        with _UniqueNames.__lock:
            _UniqueNames.__counter += 1
            return _UniqueNames.__name_prefix + str(_UniqueNames.__counter)

def _get_expr_operands(expr):
    '''
    Returns a list of the operands, from right to left, extracted from an
    expression with the following format:
        op_1 << op_2 << ... << op_n
    '''
    operands = []               # operands from right to left
    cur = expr
    while isinstance(cur, ast.BinOp) and isinstance(cur.op, ast.LShift):
        operands.append(cur.right)
        cur = cur.left
    operands.append(cur)
    return operands

def _from_load_to_store(node):
    '''
    Transforms the context (ctx) of ast.Name nodes from ast.Load to ast.Store
    starting from node and visiting ast.List and ast.Tuple nodes.
    '''
    if isinstance(node, (ast.List, ast.Tuple)):
        node.ctx = ast.Store()
        for child in node.elts:
            if not _from_load_to_store(child):
                return False
    elif isinstance(node, ast.Name):
        node.ctx = ast.Store()
    else:
        return False            # type unexpected
    return True

def _change_tree_pos(node, lineno, col_offset):
    '''
    Updates the fields lineno and col_offset of every ast node in the tree
    rooted in node with the values supplied.
    '''
    if 'lineno' in node._attributes:
        node.lineno = lineno
        node.col_offset = col_offset

    for name, field in ast.iter_fields(node):
        if isinstance(field, ast.AST):
            _change_tree_pos(field, lineno, col_offset)
        elif isinstance(field, list):
            for elem in field:
                if isinstance(elem, ast.AST):
                    _change_tree_pos(elem, lineno, col_offset)

def _get_expr_data(expr, code_info):
    '''
    Extracts the operands from an expression with one of the following formats:
        'with' [ret_vars <<] func(func_args) [<< codekw] << code_args
        'with' [ret_vars <<] func(func_args) << ':multi'
    where code_args can also be ':dict' and ':list' which are special cases. In
    those cases, codekw is allowed.

    To see the expected format for the params, see the main documentation.

    Returns:
        (ret_vars,                  # ast node or None (if not present)
         func_args,                 # ast.Call node
         codekw,                    # string or None (if not present)
         code_args)                 # ast.Str node
    '''
    operands = _get_expr_operands(expr)     # in reversed order
    ret_vars = codekw = None        # optional operands

    num_ops = len(operands)
    if num_ops < 2:
        return None                 # at least two operands expected

    # Reads code_args.
    code_args = operands[0]
    if isinstance(code_args, ast.Name) and code_args.id in ('_', 'None'):
        code_args = ast.Str('', lineno = code_args.lineno,
                            col_offset = code_args.col_offset)
    elif not isinstance(code_args, ast.Str):
        return None                 # wrong type
    next = 1
    if code_args.s != ':multi':
        # Reads codekw (if present).
        if isinstance(operands[1], ast.Str):
            codekw = operands[1].s
            next = 2
        elif isinstance(operands[1], ast.Name):
            codekw = operands[1].id
            next = 2

    # Reads call.
    if num_ops <= next or not isinstance(operands[next], ast.Call):
        return None                 # call expected
    call = operands[next]
    next += 1

    if num_ops > next + 1:
        return None                 # too many operands

    # Reads ret_vars (if present).
    if num_ops == next + 1:
        ret_vars = operands[next]
        if isinstance(ret_vars, ast.Str):
            src = ret_vars.s + ' = 0'
            try:
                assign = ast.parse(src, '<string>', 'exec').body[0]
            except Exception as e:
                _raise(type(e), e.msg, ret_vars.lineno,
                       e.offset + ret_vars.col_offset, code_info)
            target = assign.targets[0]              # name, tuple or list
            _change_tree_pos(target, ret_vars.lineno, ret_vars.col_offset)
            ret_vars = target
        elif not _from_load_to_store(ret_vars):     # modifies ret_vars
            _raise(SyntaxError, 'Invalid ret_vars.', ret_vars.lineno,
                   ret_vars.col_offset, code_info)

    return ret_vars, call, codekw, code_args

def _get_int_expr_data(expr, code_info):
    '''
    Extracts the operands from an expression with the following format:
        'with' [<< codekw] << code_args

    To see the expected format for the params, see the main documentation.

    Returns:
        (codekw,                    # string or None (if not present)
         code_args)                 # ast.Str node
    '''
    operands = _get_expr_operands(expr)     # in reversed order
    codekw = None                           # optional operand

    num_ops = len(operands)
    if num_ops not in (1, 2):
        # There must be at least one operand (Python would have raised a
        # SyntaxError otherwise).
        wrong_op = operands[num_ops - 3]    # first extra operand
        _raise(SyntaxError,
               'Internal Withs take 1 or 2 arguments ({} given).'.format(num_ops),
               wrong_op.lineno, wrong_op.col_offset, code_info)

    # Reads code_args.
    code_args = operands[0]
    if isinstance(code_args, ast.Name) and code_args.id in ('_', 'None'):
        code_args = ast.Str('', lineno = code_args.lineno,
                            col_offset = code_args.col_offset)
    elif not isinstance(code_args, ast.Str):
        _raise(SyntaxError, 'code_args must be a string, _ or None.',
               code_args.lineno, code_args.col_offset, code_info)

    # Reads codekw (if present).
    if num_ops == 2:
        if isinstance(operands[1], ast.Str):
            codekw = operands[1].s
            next = 2
        elif isinstance(operands[1], ast.Name):
            codekw = operands[1].id
            next = 2
        else:
            _raise(SyntaxError, 'codekw must be a string or an identifier.',
                   operands[1].lineno, operands[1].col_offset, code_info)

    return codekw, code_args

def _create_block_func(code_args, with_node, pos, code_info):
    '''
    Creates and returns an ast.FunctionDef node that describes the anonymous
    function
        def anon_func(code_args):
            <body>
    where <body> is the body of with_node (ast.With node).
    '''
    func_name = _UniqueNames.get_name()
    src = 'def {}({}):\n    pass\n'.format(func_name, code_args.s)
    try:
        block_func = ast.parse(src, '<string>', 'exec').body[0]
    except Exception as e:
        args_pos = src.find('(') + 1
        _raise(type(e), e.msg, code_args.lineno,
               e.offset - args_pos + code_args.col_offset, code_info)
    block_func.body = with_node.body
    block_func.lineno = pos['lineno']
    block_func.col_offset = pos['col_offset']
    return func_name, block_func

def _get_arg_src_loc_dict(call):
    '''
    Returns the source code position (a dict with 'lineno' and 'col_offset')
    that a new argument passed to the function (related to) 'call' should have.
    
    Only positional args are considered because it seems that pos information
    about keyword args is unavailable in ast trees.
    '''
    if call.args:
        return {'lineno' : call.args[-1].lineno,
                'col_offset' : call.args[-1].col_offset}
    else:
        return {'lineno' : call.lineno, 'col_offset' : call.col_offset}

def _pass_arg_to_call(arg_node, codekw, call, pos):
    '''
    Adds the argument represented by arg_node to the function represented by
    call. If codekw is None, the argument is positional, otherwise it has
    keyword codekw.
    '''
    if codekw is not None:
        call.keywords.append(ast.keyword(codekw, arg_node, **pos))
    else:
        call.args.append(arg_node)

def _pass_block_func_to_call(func_name, codekw, call, pos):
    '''
    Passes the block-function of name func_name to the function represented by
    call. If codekw is None, func_name is passed as a positional argument,
    otherwise it is passed as a keyword argument of name codekw.
    '''
    _pass_arg_to_call(ast.Name(func_name, ast.Load(), **pos), codekw, call, pos)

def _modify_body_list_with(nodes, node_idx, code_info):
    '''
    Performs the With-rewriting described in the main documentation by replacing
    nodes[node_idx : node_idx + 1] with N nodes, where N >= 2.

    Returns N (as defined above).

    nodes is a body list, i.e. a block of code (the body of a module, a function,
    a for statement, etc...).
    '''
    prev_len = len(nodes)
    node = nodes[node_idx]
    data = _get_expr_data(node.context_expr, code_info)
    if data is None:
        return 1
    ret_vars, call, codekw, code_args = data
    main_with_loc = _get_arg_src_loc_dict(call)
    if code_args.s in (':multi', ':dict', ':list'):
        mode = code_args.s

        if mode == ':dict':
            arg_node = ast.Dict([], [], **main_with_loc)
        elif mode == ':list':
            arg_node = ast.List([], ast.Load(), **main_with_loc)

        new_block_funcs = []

        assert node.body
        for i in range(len(node.body)):     # for each internal With statement
            int_with = node.body[i]
            if not isinstance(int_with, ast.With):
                _raise(SyntaxError, 'Internal With statement expected.',
                       int_with.lineno, int_with.col_offset, code_info)

            int_codekw, int_code_args = _get_int_expr_data(int_with.context_expr,
                                                           code_info)

            func_name, block_func = _create_block_func(int_code_args, int_with,
                                                       main_with_loc, code_info)
            new_block_funcs.append(block_func)

            if mode == ':dict':
                if int_codekw is None:
                    _raise(SyntaxError,
                           ':dict forbids internal Withs without codekw args.',
                           int_with.lineno, int_with.col_offset, code_info)
                arg_node.keys.append(ast.Str(int_codekw, **main_with_loc))
                arg_node.values.append(ast.Name(func_name, ast.Load(),
                                                **main_with_loc))
            elif mode == ':list':
                if int_codekw is not None:
                    _raise(SyntaxError,
                           ':list forbids internal Withs with codekw args.',
                           int_with.lineno, int_with.col_offset, code_info)
                arg_node.elts.append(ast.Name(func_name, ast.Load(),
                                              **main_with_loc))
            else:      # ':multi'
                _pass_block_func_to_call(func_name, int_codekw, call, main_with_loc)

        nodes[node_idx : node_idx] = new_block_funcs
        node_idx += len(new_block_funcs)    # updates pos of external With node

        if mode != ':multi':                # :dict or :list
            _pass_arg_to_call(arg_node, codekw, call, main_with_loc)
    else:               # single-block With statement
        func_name, block_func = _create_block_func(code_args, node, main_with_loc,
                                                   code_info)
        _pass_block_func_to_call(func_name, codekw, call, main_with_loc)

        nodes.insert(node_idx, block_func)
        node_idx += 1                       # updates pos of external With node

    # Replaces the With node with an Expr(Call) or an Assignment relative to the
    # func to call or assignment to perform, respectively.
    if ret_vars is not None:                # we have an Assignment
        assign = ast.Assign([ret_vars], call)
        assign.lineno = ret_vars.lineno
        assign.col_offset = ret_vars.col_offset
        nodes[node_idx] = assign
    else:                                   # we have just one Call
        nodes[node_idx] = ast.Expr(call, **main_with_loc)

    return 1 + len(nodes) - prev_len        # we added additional nodes

def _modify_body_list_expr(nodes, node_idx, code_info, mod_end_line):
    '''
    Performs the Expression-rewriting described in the main documentation by
    replacing nodes[node_idx : node_idx + 1] with N nodes, where N >= 2.

    Returns N (as defined above).

    nodes is a body list, i.e. a block of code (the body of a module, a function,
    a for statement, etc...).
    '''
    globals = code_info.cframe.f_globals
    node = nodes[node_idx]
    if ('value' in node._fields
            and isinstance(node.value, ast.Call)
            and 'func' in node.value._fields):
        func_node = node.value.func
        todo = None
        if isinstance(func_node, ast.Attribute):
            func_name = func_node.attr
            if not isinstance(func_node.value, ast.Subscript) \
               and not isinstance(func_node.value, ast.Attribute)\
               and not isinstance(func_node.value, ast.Call):
                mod_name = func_node.value.id
                if (mod_name in globals
                        and isinstance(globals[mod_name], ModuleType)
                        and hasattr(globals[mod_name], func_name)):
                    func = getattr(globals[mod_name], func_name)
                    if func == rewrite:
                        todo = 'del'
                    elif func == end_of_module:
                        mod_end_line[0] = node.lineno
        elif isinstance(func_node, ast.Name):
            if func_node.id in globals and globals[func_node.id] == rewrite:
                todo = 'del'
            elif func_node.id in ('_global', '_nonlocal', '_return', '_yield'):
                todo = 'mod'

        if todo == 'del':
            nodes[node_idx] = ast.Pass(lineno = node.lineno,
                                       col_offset = node.col_offset)
        elif todo == 'mod':
            arg_names = []
            if func_node.id == '_global' or func_node.id == '_nonlocal':
                for arg in node.value.args:
                    if not isinstance(arg, ast.Name):
                        _raise(SyntaxError,
                               'Identifiers expected in {}.'.format(func_node.id),
                               arg.lineno, arg.col_offset, code_info)
                    arg_names.append(arg.id)
                if func_node.id == '_global':
                    nodes[node_idx] = ast.Global(arg_names, lineno = node.lineno,
                                                 col_offset = node.col_offset)
                else:             # _nonlocal
                    nodes[node_idx] = ast.Nonlocal(arg_names, lineno = node.lineno,
                                                   col_offset = node.col_offset)
            else:
                if node.value.keywords:
                    wrong_node = node.value.keywords[0].value
                    _raise(SyntaxError,
                           '{} forbids keyword args.'.format(func_node.id),
                           wrong_node.lineno, wrong_node.col_offset, code_info)
                args = node.value.args
                loc_args = {'lineno' : node.lineno, 'col_offset' : node.col_offset}
                all_args = dict(loc_args)
                if len(args) > 1:
                    all_args['value'] = ast.Tuple(args, ast.Load(),
                                                  lineno = args[0].lineno,
                                                  col_offset = args[0].col_offset)
                elif args:       # single element
                    all_args['value'] = node.value.args[0]
                else:            # empty
                    all_args['value'] = None

                if func_node.id == '_return':
                    nodes[node_idx] = ast.Return(**all_args)
                else:            # '_yield'
                    nodes[node_idx] = ast.Expr(ast.Yield(**all_args), **loc_args)
    return 1

def _modify_body_list(nodes, node_idx, code_info, mod_end_line):
    '''
    Rewrites nodes[node_idx] using the rewriting rules described in the main
    documentation.

    Calls _modify_body_list_with or _modify_body_list_expr depending on the type
    of nodes[node_idx] and returns the value returned by them.
    '''
    if isinstance(nodes[node_idx], ast.With):
        return _modify_body_list_with(nodes, node_idx, code_info)
    else:
        return _modify_body_list_expr(nodes, node_idx, code_info, mod_end_line)

def _rewrite_tree_int(node, code_info, mod_end_line):
    ''' "Internal" function called by _rewrite_tree. '''
    for name, field in ast.iter_fields(node):
        if isinstance(field, ast.AST):
            if _rewrite_tree_int(field, code_info, mod_end_line):
                raise Exception("This was unexpected!")
        elif isinstance(field, list):
            i = 0
            while i < len(field):
                subfield = field[i]
                if isinstance(subfield, ast.AST):
                    if _rewrite_tree_int(subfield, code_info, mod_end_line):
                        i += _modify_body_list(field, i, code_info, mod_end_line) - 1
                i += 1
    return isinstance(node, (ast.With, ast.Expr))

def _rewrite_tree(tree, code_info):
    '''
    Walks the tree, find nodes that need to be rewritten and rewrite them as
    described in the main documentation.
    '''
    mod_end_line = [None]
    _rewrite_tree_int(tree, code_info, mod_end_line)
    return mod_end_line[0]

def _skip_module_code(mod_end_line, code_info):
    '''
    Skips the original (i.e. un-rewritten) code of the module.
    
    This function basically jumps to the line mod_end_line of the module.
    '''
    cframe = code_info.cframe
    old_sys_trace = sys.gettrace()
    old_cframe_trace = cframe.f_trace

    def my_trace(frame, event, arg):
        cframe.f_lineno = mod_end_line
        sys.settrace(old_sys_trace)
        cframe.f_trace = old_cframe_trace

    sys.settrace(my_trace)
    cframe.f_trace = my_trace

def end_of_module():
    '''
    Marks the last line of a module which uses our services.
    
    This is the line that _skip_module_code jumps to to skip the original code
    of the module.
    '''
    pass

def _bug_check_int(node, code_info):
    ''' "Internal" function called by _bug_check. '''
    if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
        # with statements inside of function or class defs are allowed.
        return
    if isinstance(node, ast.With):
        link = 'http://bugs.python.org/issue14612'
        msg = ('Because of a bug in Python (' + link + '), with statements '
                'appearing directly at the module level in imported modules '
                'are not allowed, for now.')
        _raise(SyntaxError, msg, node.lineno, 0, code_info)
    for name, field in ast.iter_fields(node):
        if isinstance(field, list):
            for subfield in field:
                if isinstance(subfield, ast.AST):
                    _bug_check_int(subfield, code_info)
        elif isinstance(field, ast.AST):
            _bug_check_int(field, code_info)

def _bug_check(tree, code_info):
    '''
    Makes sure that we avoid the bug described in the main doc, i.e. that there
    are not with statements at the module level inside an imported module.
    '''
    if sys.argv[0] == code_info.fname:
        return                  # we are not in an imported module
    assert isinstance(tree, ast.Module)
    _bug_check_int(tree, code_info)

class _code_info:
    '''
    Simple structure to memorize stack frame, file name and source code of the
    caller.
    '''
    def __init__(self, cframe, fname, src):
        self.cframe = cframe
        self.fname = fname
        self.src = src

def rewrite(debug=False):
    ''' Gets, modifies and executes the source code of the caller. '''

    cframe = inspect.currentframe().f_back
    fname = cframe.f_code.co_filename

    with open(fname) as f:
        src = f.read()

    code_info = _code_info(cframe, fname, src)

    tree = ast.parse(src, fname, 'exec')

    in_imported_module = sys.argv[0] != fname

    to_reraise = None
    try:
        _bug_check(tree, code_info)

        mod_end_line = _rewrite_tree(tree, code_info)
        if in_imported_module and mod_end_line is None:
            _raise(EndOfModuleError,
                   'You must call end_of_module at the end of a module.',
                   src.count('\n'), 0, code_info)
        code = compile(tree, fname, 'exec')
    except _Reraise as r:
        to_reraise = r
    except SyntaxError as s:
        to_reraise = _Reraise(type(s), s.msg, s.lineno, s.offset, code_info)
    except Exception as e:
        if debug:
            print '<<<Rewrite error'
            from traceback import print_exc
            print_exc()
            print '>>>'
        raise e
    finally:
        # To keep the stack shorter, we don't put this code in a function.
        # NOTE: We assume that the exception derives from SyntaxError
        if to_reraise is not None:
            exc_type, msg, lineno, col_offset, code_info = to_reraise.get_data()

            line_str = code_info.src.split('\n')[lineno - 1]
            if re.match(r'[ \t]*$', line_str):
                line_str = '<blank line>'

            # This forces the debuggers to break on the user code at the right
            # line, but the error message is repeated.
            #
            # Format:
            #     raise exc_type(msg, (code_info.fname, lineno, col_offset + 1,
            #                          line_str))
            vSpace = '\n' * (lineno - 1)
            hSpace = '1' * (col_offset - 2) + ';' if col_offset >= 2 else ''
            src = (vSpace + hSpace + "raise {}({!r}, ({!r}, {}, {}, {!r}))"
                   .format(exc_type.__name__, msg, fname, lineno,
                           col_offset + 1, line_str))
            exec(compile(src, code_info.fname, 'exec'))

    # Runtime errors are handled automatically.

    exec(code, cframe.f_globals, cframe.f_locals)

    if not in_imported_module:
        quit()
    else:
        _skip_module_code(mod_end_line, code_info)

start = rewrite
finish = end_of_module