# -*- coding: utf-8 -*-
"""
    flask.helpers
    ~~~~~~~~~~~~~

    Implements various helpers.

    :copyright: (c) 2011 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import with_statement

import os
import sys
import pkgutil
import posixpath
import mimetypes
from functools import update_wrapper, wraps
from time import time
from zlib import adler32
from multiprocessing import (RLock,
                             Lock,
                             Value,
                             active_children as mp_active_children)

from urlparse import urlparse, urlunparse, ParseResult
from itertools import izip_longest
import inspect
from types import StringTypes

from psutil import Process as PSProcess

from werkzeug.routing import BuildError
from werkzeug.utils import import_string
from werkzeug.urls import url_quote
from werkzeug.datastructures import ImmutableDict


from .globals import swarm

# sentinel
_missing = object()


# what separators does this operating system provide that are not a slash?
# this is used by the send_from_directory function to ensure that nobody is
# able to access files from outside the filesystem.
_os_alt_seps = list(sep for sep in [os.path.sep, os.path.altsep]
                    if sep not in (None, '/'))


def _endpoint_from_swarm_func(subprocess_func):
    """Internal helper that returns the default endpoint for a given
    function.  This always is the function name.
    """
    assert subprocess_func is not None, 'expected view func if endpoint ' \
                                    'is not provided.'
    return subprocess_func.__name__


def get_root_path(import_name):
    """Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.

    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        filepath = sys.modules[import_name].__file__

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def find_package(import_name):
    """Finds a package and returns the prefix (or None if the package is
    not installed) as well as the folder that contains the package or
    module as a tuple.  The package path returned is the module that would
    have to be added to the pythonpath in order to make it possible to
    import the module.  The prefix is the path below which a UNIX like
    folder structure exists (lib, share etc.).
    """
    root_mod_name = import_name.split('.')[0]
    loader = pkgutil.get_loader(root_mod_name)
    if loader is None or import_name == '__main__':
        # import name is not found, or interactive/main module
        package_path = os.getcwd()
    else:
        # For .egg, zipimporter does not have get_filename until Python 2.7.
        if hasattr(loader, 'get_filename'):
            filename = loader.get_filename(root_mod_name)
        elif hasattr(loader, 'archive'):
            # zipimporter's loader.archive points to the .egg or .zip
            # archive filename is dropped in call to dirname below.
            filename = loader.archive
        else:
            # At least one loader is missing both get_filename and archive:
            # Google App Engine's HardenedModulesHook
            #
            # Fall back to imports.
            __import__(import_name)
            filename = sys.modules[import_name].__file__
        package_path = os.path.abspath(os.path.dirname(filename))
        # package_path ends with __init__.py for a package
        if loader.is_package(root_mod_name):
            package_path = os.path.dirname(package_path)

    site_parent, site_folder = os.path.split(package_path)
    py_prefix = os.path.abspath(sys.prefix)
    if package_path.startswith(py_prefix):
        return py_prefix, package_path
    elif site_folder.lower() == 'site-packages':
        parent, folder = os.path.split(site_parent)
        # Windows like installations
        if folder.lower() == 'lib':
            base_dir = parent
        # UNIX like installations
        elif os.path.basename(parent).lower() == 'lib':
            base_dir = os.path.dirname(parent)
        else:
            base_dir = site_parent
        return base_dir, package_path
    return None, package_path


class locked_cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value.  Works like the one in Werkzeug but has a lock for
    thread safety.
    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func
        self.lock = RLock()

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        with self.lock:
            value = obj.__dict__.get(self.__name__, _missing)
            if value is _missing:
                value = self.func(obj)
                obj.__dict__[self.__name__] = value
            return value


class _PackageBoundObject(object):
    def __init__(self, import_name):
        #: The name of the package or module.  Do not change this once
        #: it was set by the constructor.
        self.import_name = import_name

        #: Where is the app root located?
        self.root_path = get_root_path(self.import_name)

        self._persistance_folder = None

    def _get_persistance_folder(self):
        if self._persistance_folder is not None:
            return os.path.join(self.root_path, self._persistance_folder)

    def _set_persistance_folder(self, value):
        self._persistance_folder = value
    persistance_folder = property(_get_persistance_folder,
                                  _set_persistance_folder)
    del _get_persistance_folder, _set_persistance_folder

    @property
    def has_persistance_folder(self):
        """This is `True` if the package bound object's container has a
        folder named ``'static'``.

        .. versionadded:: 0.5
        """
        return self.static_folder is not None

    def open_resource(self, resource, mode='rb'):
        """Opens a resource from the application's persistance folder.  To see
        how this works, consider the following folder structure::

            /myapplication.py
            /schema.sql
            /static
                /style.css
            /templates
                /layout.html
                /index.html

        If you want to open the `schema.sql` file you would do the
        following::

            with app.open_resource('schema.sql') as f:
                contents = f.read()
                do_something_with(contents)

        :param resource: the name of the resource.  To access resources within
                         subfolders use forward slashes as separator.
        """
        return open(os.path.join(self.root_path, resource), mode)

def caller_locals(decorators=0):
    f = sys._getframe(2+decorators)
    args = inspect.getargvalues(f)
    return args[3]

def cascade_reload(module):
    if isinstance(module, (list, tuple)):
        for m in module:
            cascade_reload(m)
        return

    reload(module)
    if hasattr(module, '__reload__'):
        for module in module.__reload__:
            cascade_reload(module)

def show_in_browser(text, coding='utf-8'):
    BROWSER = 'chromium-browser' if not 'BROWSER' in globals() else globals()['BROWSER']
    from tempfile import NamedTemporaryFile
    from subprocess import Popen
    import codecs
    from lxml.etree import tounicode
    temp = None

    with NamedTemporaryFile(delete=False) as t:
        t = codecs.getwriter(coding)(t)
        t.write(text.decode(coding))
        temp = t

    if temp:
        b = Popen('%s '%BROWSER+'file://'+temp.name, shell=True)
        b.wait()

obj_converter = lambda obj_path: import_string(obj_path) \
                                   if isinstance(obj_path, basestring)\
                                   else obj_path

def obj_list_converter(string_or_stringlist):
    if string_or_stringlist is None:
        return None
    if isinstance(string_or_stringlist, StringTypes):
        return obj_converter(string_or_stringlist)
    else:
        lst = []
        for obj_path in string_or_stringlist:
            lst.append(obj_converter(obj_path))
        return lst