import time
from copy import copy
from werkzeug import Local as wLocal, LocalStack
from os import getpid
from threading import current_thread as get_ident

try:
    from greenlet import getcurrent
    AWARE_OF_GREENLETS = True
except ImportError:
    AWARE_OF_GREENLETS = False



from psutil import Process as PSProcess, NoSuchProcess
"""
Multilayer local
keeps track of greenlet
"""

import threading
class Thread(threading.Thread):
    def __init__(self, *args, **kwargs):
        self._parent = threading.current_thread()
        super(Thread, self).__init__(*args, **kwargs)
threading.Thread = Thread

def get_iter_thread_ancestors():
    stopper_ident = get_ident()
    def iter_thread_ancestors(ident):
        while hasattr(ident, '_parent') and not ident==stopper_ident:
            ident = ident._parent
            yield ident
            
    return iter_thread_ancestors

def get_iter_process_ancestors():
    stopper_ident = getpid()
    def iter_process_ancestors(ident):
        try:
            ps = PSProcess(ident)
            while not ps.pid == stopper_ident:
                yield ps.ppid
                ps = PSProcess(ps.ppid)
        except NoSuchProcess:
            pass
    return iter_process_ancestors

if AWARE_OF_GREENLETS:
    def get_iter_green_ancestors():
        stopper_ident = getcurrent()
        def iter_green_ancestors(ident):
            while ident.parent is not None and not ident == stopper_ident:
                yield ident.parent
                ident = ident.parent
        return iter_green_ancestors

    IDENTS = (getcurrent, get_ident, getpid)

    IDENT_ANCESTOR_ITERATOR_GETTERS = (get_iter_green_ancestors,
                                       get_iter_thread_ancestors,
                                       get_iter_process_ancestors)
else:
    IDENTS = (get_ident, getpid)

    IDENT_ANCESTOR_ITERATOR_GETTERS = (get_iter_thread_ancestors,
                                       get_iter_process_ancestors)

class Local(wLocal):
    __slots__ = ('__storage__', '__ident_func__')
    def __init__(self, ident):
        object.__setattr__(self, '__storage__', {})
        object.__setattr__(self, '__ident_func__', ident)

class LocalGlobal(object):
    """
    global cant be released from another ident
    on set:
        if no such name in all locals and not such name in global - set value to global
        global can be released only from same ident ([process, glet]) combination
        global can be deleted only from same ident ([process, glet]) combination

    """
    __slots__ = (
                 '__ident_func__',
                 '__ident_funcs__',
                 '__ancestor_iterators__',
                 '__locals__',
                 )
    
    def __init__(self):
        object.__setattr__(self, '__ident_funcs__', IDENTS)
        object.__setattr__(self,
                           '__ancestor_iterators__',
                           [g() for g in IDENT_ANCESTOR_ITERATOR_GETTERS])
        
        object.__setattr__(self, '__ident_func__', 
                                 lambda:[ident() for ident in self.__ident_funcs__])
        object.__setattr__(self, '__locals__', {})

        
        for ident in self.__ident_funcs__:
            self.__locals__[ident] = Local(ident)
            
    
    def __release_local__(self):
        for ident in self.__ident_funcs__:
            self.__locals__[ident].__release_local__()

    def __get_from_ancestor__(self, name, ident):
        ancestor_iterator = \
            self.__ancestor_iterators__[self.__ident_funcs__.index(ident)]
        for ancestor in ancestor_iterator(ident()):
            if ancestor in self.__locals__[ident].__storage__ \
            and name in self.__locals__[ident].__storage__[ancestor]:
                return self.__locals__[ident].__storage__[ancestor][name]


    def __getattr__(self, name):
        for ident in self.__ident_funcs__:
            ident_value = ident()
            if ident_value in self.__locals__[ident].__storage__\
            and name in self.__locals__[ident].__storage__[ident_value]:
                return self.__locals__[ident].__storage__[ident_value][name]
            else:
                top_attr = self.__get_from_ancestor__(name, ident)
                if top_attr:
                    return top_attr
                elif ident == self.__ident_funcs__[-1]:
                    raise AttributeError(name)
    
    def __has_locally__(self, name):
        return any((hasattr(self.__locals__[ident], name) for ident in self.__ident_funcs__))

    def __setattr__(self, name, value):
        setattr(self.__locals__[self.__ident_funcs__[0]], name, value)
        #if len(self.__ident_funcs__) > 1:
        for ident in self.__ident_funcs__[1:]:
            setattr(self.__locals__[ident], name, value)

    def __delattr__(self, name):
        for ident in self.__ident_funcs__:
            delattr(self.__locals__[ident], name)

        
class LocalGlobalStack(LocalStack):
    def __init__(self):
        self._local = LocalGlobal()

    def ensure_stack_is_local(self):
        if not self._local.__has_locally__('stack'):
            try:
                self._local.stack = copy(getattr(self._local, 'stack'))
            except AttributeError:
                self._local.stack = []

    def push(self, value):
        self.ensure_stack_is_local()
        return super(LocalGlobalStack, self).push(value)

    def pop(self):
        self.ensure_stack_is_local()
        return super(LocalGlobalStack, self).pop()

    def __len__(self):
        try:
            return len(self._local.stack)
        except AttributeError:
            return 0
    
    @property
    def top(self):
        """
           renamed top to bottom - when using [with codeblocks] statements
           topmost element is at start
        """
        try:
            return self._local.stack[0]
        except (AttributeError, IndexError):
            return None

    @property
    def bottom(self):
        """
           renamed top to bottom - when using [with codeblocks] statements
           topmost element is at start
        """
        try:
            return self._local.stack[-1]
        except (AttributeError, IndexError):
            return None


if __name__ == '__main__':
    import unittest
    from multiprocessing import Process
    # from gogreen.coro import Thread as GreenThread
    class TestLocalGlobal(unittest.TestCase):
        def test_process(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')

        def test_process_sync_subprocess(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                def test_subprocess():
                    self.assertTrue(local.name == 'stored_in_process')
                    local.name = 'stored_in_sync_subprocess'
                    self.assertTrue(local.name == 'stored_in_sync_subprocess')
                    self.assertTrue(local.g == 'global_only')

                p = Process(target=test_subprocess,)
                p.start()
                p.join()

                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')

        def test_process_async_subprocess(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                def test_subprocess():
                    self.assertTrue(local.name == 'stored_in_process')
                    local.name = 'stored_in_async_subprocess'
                    self.assertTrue(local.name == 'stored_in_async_subprocess')
                    self.assertTrue(local.g == 'global_only')

                p = Process(target=test_subprocess,)
                p.start()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')
            time.sleep(.2)

        def test_sync_process_async_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                case = self
                local.name = 'stored_in_process'
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                thread = TestThread()
                thread.start()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')
                time.sleep(.2)

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')

        def test_async_process_async_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                case = self
                local.name = 'stored_in_process'
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                thread = TestThread()
                thread.start()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')
                time.sleep(.2)

            p = Process(target=test_process,)
            p.start()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')
            time.sleep(.4)

        def test_async_process_sync_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                case = self
                local.name = 'stored_in_process'
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                thread = TestThread()
                thread.start()
                thread.join()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')
            time.sleep(.4)            
        
        def test_sync_process_sync_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                case = self
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                thread = TestThread()
                thread.start()
                thread.join()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')

        def test_sync_process_sync_subthread_sync_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                case = self
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')

                        class TestSubThread(threading.Thread):
                            def run(self):
                                case.assertTrue(local.name == 'stored_in_subthread')
                                local.name = 'stored_in_subsubthread'
                                case.assertTrue(local.name == 'stored_in_subsubthread')
                                case.assertTrue(local.g == 'global_only')
                        thread = TestSubThread()
                        thread.start()
                        thread.join()
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')

                thread = TestThread()
                thread.start()
                thread.join()
                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')

        def test_async_process_async_subthread_async_subthread(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                case = self
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')

                        class TestSubThread(threading.Thread):
                            def run(self):
                                case.assertTrue(local.name == 'stored_in_subthread')
                                local.name = 'stored_in_subsubthread'
                                case.assertTrue(local.name == 'stored_in_subsubthread')
                                case.assertTrue(local.g == 'global_only')
                        thread = TestSubThread()
                        thread.start()
                        
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                        time.sleep(0.2)

                thread = TestThread()
                thread.start()

                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')
                time.sleep(0.3)

            p = Process(target=test_process,)
            p.start()
            
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')
            time.sleep(0.4)

        def test_sync_process_sync_subthread_sync_subprocess(self):
            local = LocalGlobal()
            def test_process():
                self.assertTrue(local.list == ['stored_in_global',])
                local.list.append('stored_in_process')
                case = self
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.list == ['stored_in_global',
                                                       'stored_in_process'])
                        local.list.append('stored_in_subthread')
                        case.assertTrue(local.list == ['stored_in_global',
                                                       'stored_in_process',
                                                       'stored_in_subthread'])
                        case.assertTrue(local.g == 'global_only')
                        def test_subprocess():
                            case.assertTrue(local.list == ['stored_in_global',
                                                           'stored_in_process',
                                                           'stored_in_subthread'])
                            local.list.append('stored_in_subprocess')
                            case.assertTrue(local.list == ['stored_in_global',
                                                           'stored_in_process',
                                                           'stored_in_subthread',
                                                           'stored_in_subprocess'])
                            case.assertTrue(local.g == 'global_only')
                        p = Process(target=test_subprocess)
                        p.start()
                        p.join()
                        case.assertTrue(local.list == ['stored_in_global',
                                                       'stored_in_process',
                                                       'stored_in_subthread'])
                        case.assertTrue(local.g == 'global_only')

                thread = TestThread()
                thread.start()
                thread.join()
                self.assertTrue(local.list == ['stored_in_global',
                                               'stored_in_process',
                                               'stored_in_subthread'])#last line is ???
                self.assertTrue(local.g == 'global_only')
            local.list = ['stored_in_global',]
            local.g = 'global_only'

            p = Process(target=test_process,)
            p.start()
            p.join()
            self.assertTrue(local.list == ['stored_in_global',])
            self.assertTrue(local.g == 'global_only')

        def test_async_process_async_subthread_async_subprocess(self):
            local = LocalGlobal()
            local.name = 'stored_in_global'
            local.g = 'global_only'
            def test_process():
                self.assertTrue(local.name == 'stored_in_global')
                local.name = 'stored_in_process'
                case = self
                class TestThread(threading.Thread):
                    def run(self):
                        case.assertTrue(local.name == 'stored_in_process')
                        local.name = 'stored_in_subthread'
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                        def test_subprocess():
                            case.assertTrue(local.name == 'stored_in_subthread')
                            local.name = 'stored_in_subprocess'
                            case.assertTrue(local.name == 'stored_in_subprocess')
                            case.assertTrue(local.g == 'global_only')
                        p = Process(target=test_subprocess)
                        p.start()
                        
                        case.assertTrue(local.name == 'stored_in_subthread')
                        case.assertTrue(local.g == 'global_only')
                        time.sleep(0.2)

                thread = TestThread()
                thread.start()

                self.assertTrue(local.name == 'stored_in_process')
                self.assertTrue(local.g == 'global_only')
                time.sleep(0.3)

            p = Process(target=test_process,)
            p.start()
            
            self.assertTrue(local.name == 'stored_in_global')
            self.assertTrue(local.g == 'global_only')
            time.sleep(0.4)

if __name__ == '__main__':
    unittest.main(verbosity=3)