# -*- coding: utf-8 -*-
"""
    flask.logging
    ~~~~~~~~~~~~~

    Implements the logging support for Flask.

    :copyright: (c) 2011 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import sys
from logging import (getLogger,
                     basicConfig,
                     DEBUG,
                     INFO,
                     ERROR, 
                     StreamHandler,
                     Handler,
                     getLoggerClass,
                     setLoggerClass,
                     Formatter)

import multiprocessing
import threading

from .helpers import locked_cached_property
from werkzeug.utils import cached_property

import traceback


import threading
ident = lambda:(multiprocessing.current_process().pid, id(threading.current_thread()))
from pprint import pprint
from traceback import print_exc
import time


class AppLogger(getLoggerClass()):
    
    @cached_property
    def recievers(self):
        return []

    def recieve(self, queue):
        reciever = threading.Thread(name='LogReciever', target=self._receive, args=(queue,))
        reciever.daemon = True
        reciever.start()
        self.recievers.append((reciever, queue))

    def _receive(self, queue):
        while True:
            try:
                record = queue.get()
                if record is None:
                    break
                for handler in self.handlers:
                    handler.emit(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except:
                traceback.print_exc(file=sys.stderr)
    
    def close(self):
        for reciever, queue in self.recievers:
            queue.put(None)
            reciever.join()

        del self.__dict__['recievers']

        for handler in self.handlers:
            handler.close()
            self.removeHandler(handler)



def create_app_logger(app):
    try:
        orig_logger_class = getLoggerClass()
        setLoggerClass(AppLogger)
        
        name = app.logger_name if app.logger_name is not None else app.import_name
        logger = getLogger(name)
        logger.setLevel(app.log_level)
        handler = app.log_handler(app.log_level)
        handler.set_name(name)
        formatter = Formatter(app.log_format,
                              datefmt=app.config.get('LOG_DATE_FORMAT', None))

        handler.setFormatter(formatter)

        logger.addHandler(handler)
        return logger
    finally:
        setLoggerClass(orig_logger_class)


def create_context_logger(context, queue):
    app = context.object
    name = app.logger_name if app.logger_name is not None else app.import_name
    name = '%s.%s'%(name, context.block_name)
    logger = getLogger(name)

    logger.setLevel(app.log_level)
    if logger.handlers:
        return logger

    handler = ContextLogHandler(queue, app.log_level)
    handler.set_name(name)
    formatter = Formatter(app.log_format,
                          datefmt=app.config.get('LOG_DATE_FORMAT', None))
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


class ContextLogHandler(Handler):
    def __init__(self,
                queue,
                *args, 
                **kwargs):
        
        self.queue = queue
        super(ContextLogHandler, self).__init__(*args, **kwargs)
        
    def send(self, message):
        self.queue.put_nowait(message)

    def _format_record(self, record):
        # ensure that exc_info and args
        # have been stringified.  Removes any chance of
        # unpickleable things inside and possibly reduces
        # message size sent over the pipe
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            dummy = self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)